"""MongoDB Connector"""

from time import sleep
from typing import Union, List, Dict
import six

from mongoengine import connect
from mongoengine.base import BaseDocument
from mongoengine.errors import MongoEngineException

from swifty.logging.logger import swifty_logger


def init_mongodb_connections(database_pool: Dict):
    """Init mongo connections

    Args:
        database_pool (dict): _description_
    """
    for alias, database in (database_pool or {}).items():
        swifty_logger.info(
            f"[SwiftyMongoConnection] - Establish connection to alias {alias}",
            database=database,
        )
        connect_to_db(alias=alias, **database)


def connect_to_db(alias: str, count: int = 3, **kwargs) -> None:  # type: ignore
    """
    Context manager to connect to the MongoDB database with retry logic.

    Args:
        alias (str): The alias for the connection.
        count (int, optional): Number of retry attempts. Defaults to 3.

    Raises:
        MongoEngineException: If unable to connect after specified attempts.
    """
    retry_count = 0
    exception = None
    connected = False

    while retry_count < count:
        try:
            # Attempt to connect
            connect(alias=alias, **kwargs)
            connected = True
            break  # Exit retry loop on successful connection
        except (MongoEngineException, ConnectionError) as ex:
            retry_count += 1
            exception = six.text_type(ex)
            sleep(0.3)  # Wait 300ms before retrying

    if not connected:
        raise MongoEngineException(
            f"Unable to connect to database after {count} attempts, specific error: {exception}"
        )


class MongoConnector:
    """MongoDB Connector class for handling database operations.

    Args:
        SwiftyLoggerMixin (_type_): _description_
    """

    def __init__(self, mongo_collection: BaseDocument) -> None:
        if not mongo_collection or not issubclass(mongo_collection, BaseDocument):
            raise MongoEngineException("Mongo collection is not defined")
        self.mongo_collection = mongo_collection

    @property
    def mongo_objects(self):
        """Get the objects manager for the MongoDB collection."""
        return getattr(self.mongo_collection, "objects")

    def insert(self, data: dict) -> None:
        """Insert a single document into the MongoDB collection.

        Args:
            data (dict): The document data to insert.
        """
        self.mongo_objects.insert(self.mongo_collection(**data))

    def insert_many(self, list_of_data: List[Dict]) -> None:
        """Insert multiple documents into the MongoDB collection.

        Args:
            list_of_data (List[Dict]): A list of document data to insert.
        """
        self.mongo_objects.insert(
            [self.mongo_collection(**data) for data in list_of_data],
        )

    def update(
        self,
        data: dict,
        filters: Union[dict, None] = None,
        q_combine: Union[None, str] = None,
        upsert: bool = True,  # ✅ Allow insert if not found
    ) -> None:
        """Update documents in the MongoDB collection.

        Args:
            data (dict): The data to update.
            filters (Union[dict, None], optional): Filters to apply. Defaults to None.
            q_combine (Union[None, str], optional): Query combination. Defaults to None.
        """
        self.mongo_objects(q_combine, **(filters or {})).update(
            __raw__={"$set": data}, upsert=upsert
        )

    def update_many(self, list_of_data: List[Dict]) -> None:
        """Replace multiple documents in the MongoDB collection.

        Args:
            list_of_data (List[Dict]): A list of document data to replace.
        """
        self.mongo_objects.replace(
            [self.mongo_collection(**data) for data in list_of_data],
        )

    def aggregate(self, pipeline: List[dict]) -> List[dict]:
        """Perform aggregation on the MongoDB collection.

        Args:
            pipeline (List[dict]): The aggregation pipeline.

        Returns:
            List[dict]: The results of the aggregation.
        """
        return self.mongo_objects.aggregate(pipeline)

    def __find(
        self,
        filters: Union[dict, None] = None,
        q_combine: Union[None, str] = None,
        order_by: str = "-version",
    ) -> List[dict]:
        """Find documents in the MongoDB collection.

        Args:
            filters (Union[dict, None], optional): Filters to apply. Defaults to None.
            q_combine (Union[None, str], optional): Query combination. Defaults to None.
            order_by (str, optional): Field to order by. Defaults to "-version".

        Returns:
            List[dict]: The found documents.
        """
        return self.mongo_objects.filter(q_combine, **(filters or {})).order_by(
            order_by
        )

    def find_one(
        self,
        filters: Union[dict, None] = None,
        q_combine: Union[None, str] = None,
        order_by: str = "-version",
    ) -> Union[None, dict]:
        """Find a single document in the MongoDB collection.

        Args:
            filters (Union[dict, None], optional): Filters to apply. Defaults to None.
            q_combine (Union[None, str], optional): Query combination. Defaults to None.
            order_by (str, optional): Field to order by. Defaults to "-version".

        Returns:
            Union[None, dict]: The found document or None if not found.
        """
        return self.__find(
            filters=filters, q_combine=q_combine, order_by=order_by
        ).first()

    def find_many(
        self,
        filters: Union[dict, None] = None,
        q_combine: Union[None, str] = None,
        order_by: str = "-version",
    ) -> List[dict]:
        """Find multiple documents in the MongoDB collection.

        Args:
            filters (Union[dict, None], optional): Filters to apply. Defaults to None.
            q_combine (Union[None, str], optional): Query combination. Defaults to None.
            order_by (str, optional): Field to order by. Defaults to "-version".

        Returns:
            List[dict]: A list of found documents.
        """
        return self.__find(
            filters=filters, q_combine=q_combine, order_by=order_by
        ).all()



"""Constants"""

from dataclasses import dataclass


@dataclass
class LogEvents:
    """_summary_"""

    ROTATED_NEW_TOKEN = "ROTATED_NEW_TOKEN"
    PERMISSIONS_DENIED_BY_ERROR = "PERMISSIONS_DENIED_BY_ERROR"


JWT_TOKEN = "jwt_token"
REFRESH_TOKEN = "refresh"
ACCESS_TOKEN = "access"



# -*- coding: utf-8 -*-
import math


ALLOWED_NAMES = {k: v for k, v in math.__dict__.items() if not k.startswith("__")}
ALLOWED_BUILTINS = {
    "_getattr_": getattr,
    "_getiter_": iter,
    "max": max,
    "min": min,
    "abs": abs,
    "len": len,
    "sum": sum,
    "any": any,
    "all": all,
    "sorted": sorted,
    "round": round,
}


class ExpressionatorException(Exception):
    """
    _summary_

    Args:
        Exception (_type_): _description_
    """


class ExpressionatorNamespace:
    """
    _summary_

    Raises:
        ExpressionatorException: _description_

    Returns:
        _type_: _description_
    """

    name = "_expr_"

    @property
    def safe_expr_namespace(self):
        """
        _summary_
        """
        return {
            func_name.replace(self.name, ""): getattr(self, func_name)
            for func_name in dir(self)
            if func_name.startswith(self.name)
        }



from copy import deepcopy
from typing import Any, Dict, List, Type
from rest_framework import serializers
from swifty.serializers.dynamic_serializer import DynamicSerializer
from swifty.serializers.fields import (
    BasicFieldSerializer,
    FIELD_TYPE_MAP,
)


def create_field_kwargs(
    field_info,
    field_type_map: Dict[str, BasicFieldSerializer] = deepcopy(FIELD_TYPE_MAP),
):
    """_summary_

    Args:
        field_info (_type_): _description_
        field_type_map (_type_): _description_

    Returns:
        _type_: _description_
    """
    field_base = field_type_map.get(field_info.get("type", "CharField"))

    if not issubclass(field_base, BasicFieldSerializer):
        return field_base, field_info

    allowed_args = field_base.static_args + field_base.extra_args
    allowed_kwargs = {}
    extra_kwargs = {}
    for k, v in field_info.items():
        if k in allowed_args:
            allowed_kwargs.update({k: v})
        else:
            extra_kwargs.update({k: v})

    return field_base, {"extra_kwargs": extra_kwargs, **allowed_kwargs}


def create_dynamic_serializer(
    fields_config: List[Dict[str, Any]],
    base_class: Type[DynamicSerializer] = DynamicSerializer,
    field_type_map: Dict[str, BasicFieldSerializer] = deepcopy(FIELD_TYPE_MAP),
) -> Type[DynamicSerializer]:
    """Create a dynamic serializer class based on field configuration.

    Args:
        field_config (List[Dict[str, Any]]): A list of dictionaries defining the fields.

    Returns:
        Type[DynamicSerializer]: A dynamically created serializer class.
    """
    fields: Dict[str, serializers.Field] = {}

    for field_info in fields_config:
        field_base, field_kwargs = create_field_kwargs(field_info, field_type_map)
        fields[field_info["field_name"]] = field_base(**field_kwargs)

    return type("DynamicSerializer", (base_class,), fields)



"""Token Manager"""

from rest_framework.exceptions import AuthenticationFailed
from rest_framework_simplejwt.tokens import AccessToken, RefreshToken
from rest_framework_simplejwt.exceptions import InvalidToken, TokenError
from swifty.auth.constants import LogEvents, ACCESS_TOKEN, REFRESH_TOKEN
from swifty.logging.logger import SwiftyLoggerMixin


class AccessTokenManager(SwiftyLoggerMixin):
    """Manages JWT access token operations including rotation and validation"""

    def __init__(self, token_data: dict[str, str]) -> None:
        self.token_data = token_data

    def rotate_new_token(self, refresh_token: str) -> str:
        """Generates new access token using refresh token

        Args:
            refresh_token: JWT refresh token string
        Returns:
            New access token string
        """
        self.token_data[ACCESS_TOKEN] = str(RefreshToken(refresh_token).access_token)
        self.logger.info(LogEvents.ROTATED_NEW_TOKEN)
        return self.token_data[ACCESS_TOKEN]

    @property
    def access_token(self) -> str:
        """Validates and returns access token, rotating if necessary

        Raises:
            AuthenticationFailed: When access token is missing
            TokenError: When token validation fails
        """
        if not (token := self.token_data.get(ACCESS_TOKEN)):
            raise AuthenticationFailed

        try:
            return str(AccessToken(token=token, verify=True))
        except (TokenError, InvalidToken):
            return self.rotate_new_token(self.token_data[REFRESH_TOKEN])



"""Permissions"""

from typing import Any, Dict, List, Optional, Union, Callable
from django.contrib.auth.models import AnonymousUser
from rest_framework.exceptions import PermissionDenied
from rest_framework.permissions import BasePermission
from rest_framework.request import Request
from rest_framework.viewsets import ViewSet
from swifty.logging.logger import SwiftyLoggerMixin
from swifty.utils.mapper import get_nested_attributes
from swifty.auth.constants import LogEvents


class SwiftyPermission(BasePermission, SwiftyLoggerMixin):
    """Permission class that checks user attributes against allowed values."""

    permission_layers: Optional[List[Dict[str, Union[str, List[Any], Callable]]]] = None

    def has_permission(self, request: Request, view: ViewSet) -> bool:
        """Check if user has required permissions based on permission_layers."""
        try:
            if isinstance(request.user, AnonymousUser):
                raise PermissionDenied

            for layer in self.permission_layers or []:
                if (
                    not (path := layer.get("path"))
                    or not (allowed := layer.get("allowed"))
                    or allowed is True
                ):
                    continue

                allowed = allowed(request, view) if callable(allowed) else allowed
                if get_nested_attributes(root=request.user, path=path) not in allowed:
                    raise PermissionDenied

            return True

        except (PermissionDenied, AttributeError, TypeError) as error:
            self.logger.error(LogEvents.PERMISSIONS_DENIED_BY_ERROR, error=error)
            return False



"""SwiftyLoggerMixin"""

import structlog
from django.conf import settings
from swifty.utils.mapper import getpath


class SwiftyLoggerMixin:
    """Mixin for structured logging."""

    logging_context: dict | None = None

    @property
    def logger(self) -> structlog.BoundLogger:
        """Get the logger instance."""
        name = f"{self.__module__}.{self.__class__.__name__}"
        _logger = structlog.getLogger(name, exc_info=settings.DEBUG)
        if self.logging_context:
            _logger.bind(**(self.logging_context or {}))
        return _logger

    def init_context(self) -> dict:
        """Initialize logging context."""
        return self.logging_context or {}

    @property
    def context(self) -> dict:
        """Get the current logging context."""
        return (
            getpath(self.logger.bind(), "_context._dict") or self.logging_context or {}
        )

    def set_extra_context(self, extra_context: dict) -> None:
        """Bind extra context to the logger."""
        self.logger.bind(**extra_context)


swifty_logger = structlog.get_logger(__name__, exc_info=settings.DEBUG)



# -*- coding: utf-8 -*-
"""Redis cache implementation for Swifty providing caching functionality with Redis backend."""

import base64

import os
import pickle
from io import BytesIO
from functools import wraps
from dataclasses import dataclass
from typing import Any, Callable, Dict, Optional, Tuple, Union
import redis

from django.conf import settings
from django_auxilium.utils.functools import BaseCache, NotInCache

from swifty.redis.constants import RedisCacheEvents
from swifty.logging.logger import SwiftyLoggerMixin
from swifty.utils.mapper import getpath


@dataclass
class ClearCacheScope:
    """Defines cache clearing scopes."""

    METHOD = "METHOD"
    PATTERN = "PATTERN"
    ALL = "ALL"


def _redis_reconnect(func: Callable) -> Callable:
    """
    Decorator to automatically reconnect Redis connection on errors.

    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except (redis.ConnectionError, redis.TimeoutError):
            self.logger.error(RedisCacheEvents.TIMEOUT_CONNECTION, exc_info=True)
            self.reconnect()
            return func(self, *args, **kwargs)

    return wrapper


class RedisCache(BaseCache, SwiftyLoggerMixin):
    """Redis cache implementation with logging and connection management."""

    DEFAULT_REDIS_CACHE_DB: int = int(os.environ.get("DEFAULT_REDIS_CACHE_DB", 1))
    DEFAULT_REDIS_CACHE_PATTERN: Dict[str, Union[str, int]] = {
        "prefix": "DEFAULT",
        "ttl": 3600,
        "vary_by": [""],
        "db": DEFAULT_REDIS_CACHE_DB,
    }
    REDIS_SESSION_CACHE: str = getattr(
        settings, "REDIS_SESSION_CACHE", "django.contrib.sessions.cache"
    )

    @classmethod
    def redis_connection(cls, pool: redis.ConnectionPool) -> redis.StrictRedis:
        """Establish a Redis connection using the provided connection pool."""
        return redis.StrictRedis(connection_pool=pool)

    @classmethod
    def connect(cls, **kwargs) -> redis.StrictRedis:
        """Get the Redis connection, creating it if necessary."""

        cache_db = getpath(kwargs, "cache_db", cls.DEFAULT_REDIS_CACHE_DB)
        cls.redis = cls.redis_connection(
            pool=settings.REDIS_CONNECTION_POOL.get(f"pool_{cache_db}")
        )
        return cls.redis

    @classmethod
    def close(cls) -> None:
        """Close the Redis connection if it exists."""

        if hasattr(cls, "redis"):
            cls.redis.close()

    @classmethod
    def reconnect(cls) -> redis.StrictRedis:
        """Reconnect to Redis, disconnecting the old connection."""
        if hasattr(cls, "redis"):
            cls.redis.connection_pool.disconnect()
            del cls.redis
        return cls.connect()

    def __init__(
        self,
        parent: Any,
        attr: str,
        ttl: Optional[int] = None,
        static_key: Optional[str] = None,
        vary_by: Tuple[str, ...] = (),
        context_getter: Optional[Callable] = None,
    ) -> None:
        super().__init__(parent, attr)
        self.ttl = ttl
        self.static_key = static_key
        self._vary_by = vary_by
        self.context_getter = context_getter

    def __contains__(self, item: str) -> bool:
        return getpath(self, item, False)

    def add(self, *args: Any, **kwargs: Any) -> Any:
        """Add a value to the cache."""
        return self.set(*args, **kwargs)

    def _get_key(self, *args: Any, **kwargs: Any) -> str:
        """Generate the cache key."""
        return self.static_key or self._generate_key(*args, **kwargs)

    def _generate_key_prefix(self) -> str:
        """Generate the key prefix based on environment and pattern."""

        pattern = self._get_parent_pattern()
        env_prefix = os.environ.get("ENVIRONMENT_NAME")
        return "_".join(filter(None, [env_prefix, pattern["prefix"]]))

    def _get_parent_pattern(self) -> Dict[str, Union[str, int]]:
        """Get the cache pattern from the parent object."""
        return getpath(
            self.parent, "redis_cache_pattern", self.DEFAULT_REDIS_CACHE_PATTERN
        )

    def _generate_key(self, *args: Any, **kwargs: Any) -> str:
        """Generate a unique cache key based on various parameters."""
        parent_name = self.parent.__class__.__name__
        vary_by = self._get_vary_by()
        parameters = (
            base64.b64encode(
                bytes(str(args + (sorted(kwargs.items()),)), encoding="utf-8")
            ).decode("utf-8")
            if args or kwargs
            else None
        )
        return "_".join(
            filter(
                None,
                [
                    self._generate_key_prefix(),
                    parent_name,
                    self.attr,
                    vary_by,
                    parameters,
                ],
            )
        )

    def _get_vary_by(self) -> str:
        """Construct the vary by string from context."""
        context = self.context
        if isinstance(context, str):
            return context
        return (
            "_".join(str(context.get(key)) for key in self.vary_by if context.get(key))
            if context and self.vary_by
            else self.logger.warn(
                f"{RedisCacheEvents.NO_CONTEXT_FOUND} {self.parent.__class__.__name__}"
            )
        )

    @property
    def vary_by(self) -> Tuple[str, ...]:
        """Get the vary by cache string."""
        return self._vary_by or self._get_parent_pattern().get("vary_by")

    @property
    def context(self) -> Any:
        """Get context from the parent object."""
        if callable(self.context_getter):
            return self.context_getter(self.parent)
        return (
            getattr(self.parent, self.context_getter, {})
            if isinstance(self.context_getter, str)
            else getattr(self.parent, "context", {})
        )

    def get(self, *args: Any, **kwargs: Any) -> Any:
        """Retrieve a value from the cache."""
        key = self._get_key(*args, **kwargs)
        try:
            value = self._get(key)
            return pickle.load(BytesIO(value))  # type: ignore
        except NotInCache:
            raise
        except Exception as error:
            self.logger.error(
                event_name=RedisCacheEvents.GET_CACHE_ERROR,
                key=key or "<could not compute cache key>",
                method=self.attr,
                parent=self.parent.__class__.__name__,
                error=error,
                exc_info=True,
            )
            raise NotInCache from error

    def set(self, value: Any, *args: Any, **kwargs: Any) -> Any:
        """Store a value in the cache after pickling it."""
        key = self._get_key(*args, **kwargs)
        encoded = pickle.dumps(value)
        self._set(key, encoded)
        return value

    def delete(self, *args: Any, **kwargs: Any) -> Optional[Any]:
        """Delete a value from the cache."""
        if (args, kwargs) == ((ClearCacheScope.METHOD,), {}):
            return self._delete_method()
        if (args, kwargs) == ((ClearCacheScope.PATTERN,), {}):
            raise ValueError("Cannot pop cache pattern from instance methods")
        if (args, kwargs) == ((ClearCacheScope.ALL,), {}):
            raise ValueError("Cannot clear cache from instance methods")
        key = self._get_key(*args, **kwargs)
        return self._delete_value(key)

    @_redis_reconnect
    def _get(self, key: str) -> bytes:
        """Retrieve a raw value from Redis."""
        connection = self.connect(
            cache_db=(
                self.parent.redis_cache_pattern.get("db", self.DEFAULT_REDIS_CACHE_DB)
                if hasattr(self.parent, "redis_cache_pattern")
                else 1
            )
        )
        value = connection.get(key)
        if value is None:
            raise NotInCache
        return value

    @_redis_reconnect
    def _set(self, key: str, value: bytes) -> None:
        """Store a raw value in Redis."""
        connection = self.connect(
            cache_db=getpath(
                self.parent,
                "redis_cache_pattern.db",
                self.DEFAULT_REDIS_CACHE_PATTERN["db"],
            )
        )
        ttl = self.ttl or self.DEFAULT_REDIS_CACHE_PATTERN["ttl"]
        connection.set(key, value)
        if ttl:
            connection.expire(key, ttl)

    @_redis_reconnect
    def _delete_value(self, key: str) -> None:
        """Delete a value from Redis."""
        connection = self.connect(
            cache_db=(
                self.parent.redis_cache_pattern.get("db", self.DEFAULT_REDIS_CACHE_DB)
                if hasattr(self.parent, "redis_cache_pattern")
                else self.DEFAULT_REDIS_CACHE_DB
            )
        )
        for full_key in connection.scan_iter(key):
            connection.delete(full_key)

    @_redis_reconnect
    def _delete_method(self) -> None:
        """Delete cache entries for a specific method."""
        connection = self.connect(
            cache_db=(
                self.parent.redis_cache_pattern.get("db", self.DEFAULT_REDIS_CACHE_DB)
                if hasattr(self.parent, "redis_cache_pattern")
                else self.DEFAULT_REDIS_CACHE_DB
            )
        )
        full_pattern = (
            "_".join(
                [
                    os.environ.get("ENVIRONMENT_NAME", ""),
                    self._get_parent_pattern()["prefix"],
                    self.parent.__class__.__name__,
                    self.attr,
                ]
            )
            + "*"
        )

        for key in connection.scan_iter(full_pattern):
            connection.delete(key)



"""Constants"""

from dataclasses import dataclass


@dataclass
class RedisCacheEvents:
    """_summary_"""

    REDIS_CACHE = "REDIS_CACHE"
    CACHE_ERROR = f"{REDIS_CACHE}: Cache error"
    TIMEOUT_CONNECTION = f"{REDIS_CACHE}: Redis connection timed out. Reconnecting."
    ERROR_CONNECTION = f"{REDIS_CACHE}: Error connecting from redis"
    ERROR_DISCONNECTING = f"{REDIS_CACHE}: Error disconnecting from redis"
    NO_CONTEXT_FOUND = f"{REDIS_CACHE}: No context found for:"
    GET_CACHE_ERROR = f"{REDIS_CACHE}: Could not get cache value"
    UNPICKLE_CACHE_ERROR = f"{REDIS_CACHE}: Could not unpickle cache value"
    SET_CACHE_ERROR = f"{REDIS_CACHE}: Could not set value in cache"
    POP_METHOD_CACHE_ERROR = f"{REDIS_CACHE}: Could not pop method cache values"
    POP_CACHE_ERROR = f"{REDIS_CACHE}: Could not pop cache value"



"""Utility module for mapping and accessing nested data structures through path expressions."""

from typing import Any, Dict, List, Optional, Union, Generator
import re
import six

PATH_DELIMITERS = [",", ".", "[", "]"]


def parse_path(path: Union[str, List[str]]) -> Generator[str, None, None]:
    """Parse a path string or list into individual keys.
    Args:
        path (Union[str, List[str]]): The path to parse.
    Yields:
        str: Individual keys from the path.
    """

    if isinstance(path, six.string_types):
        path = _split_and_trim(path, *PATH_DELIMITERS)
    for key in path:
        if isinstance(key, six.string_types):
            yield from _split_and_trim(key, *PATH_DELIMITERS)
        else:
            yield key


def _split_and_trim(string: str, *delimiters: str) -> List[str]:
    """Split a string by given delimiters and trim empty results.
    Args:
        string (str): The string to split.
        *delimiters (str): Delimiters to use for splitting.
    Returns:
        List[str]: List of non-empty substrings.
    """
    pattern = "|".join(map(re.escape, delimiters))
    return list(filter(None, re.split(pattern, string)))


def getpath(
    obj: Union[Dict, List, Any],
    path: Union[str, List[str]],
    default: Optional[Any] = None,
) -> Any:
    """Retrieve a value from a nested structure using a path.
    Args:
        obj (Union[Dict, List, Any]): The object to search.
        path (Union[str, List[str]]): The path to follow.
        default (Optional[Any]): The value to return if the path does not exist.
    Returns:
        Any: The value found at the path or the default value.
    """

    for key in parse_path(path):
        try:
            obj = getattr(obj, key) if hasattr(obj, key) else obj[key]
        except (KeyError, IndexError, TypeError, AttributeError):
            return default
        if obj is None:
            break
    return obj


def get_nested_attributes(root: Any, path: str, default: Optional[Any] = None) -> Any:
    """Get nested attributes from an object using a dot-separated path.
    Args:
        root (Any): The root object to search.
        path (str): The dot-separated path to follow.
        default (Optional[Any]): The value to return if the path does not exist.
    Returns:
        Any: The value found at the path or the default value.
    """
    try:
        return six.moves.reduce(getattr, [root] + path.split("."))
    except AttributeError:
        return default



"""Constants"""

from dataclasses import dataclass


@dataclass
class SwiftyCacheEvents:
    """_summary_"""

    SWIFTY_CACHE = "SWIFTY_CACHE"
    CACHED_DATA = f"{SWIFTY_CACHE}: Cached data was get"
    CREATE_NEW_CACHE = f"{SWIFTY_CACHE}: Create new cache"
    DATA_NOT_IN_CACHE = f"{SWIFTY_CACHE}: Data not in cache"



"""Module for dynamic serializers using Django REST Framework."""

import json
from copy import deepcopy
from typing import Any, Dict

from rest_framework import serializers

from swifty.expressionator.manager import ExpressionatorManager
from swifty.expressionator.namespace import ExpressionatorNamespace
from swifty.utils.helpers import (
    is_any_nested_item_in_list,
    map_calculated_data,
    get_fields_config,
    json_handler,
)


class DynamicSerializer(serializers.Serializer):
    """Base class for dynamic serializers."""

    namespace = ExpressionatorNamespace
    calculators = None
    pk_field = None
    is_update = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_calcualtors()
        data = kwargs.get("data")
        if isinstance(data, dict) and "pk_field" in data:
            self.pk_field = kwargs.get("data").get("pk_field")

    def expressionator(self, data=None):
        """_summary_

        Args:
            data (_type_): _description_

        Returns:
            _type_: _description_
        """
        return ExpressionatorManager(data=data, namespace=self.namespace)

    def create(self, *args, **kwargs) -> Any:
        """Create an instance based on validated data."""

    def update(self, *args, **kwargs) -> Any:
        """Update an instance based on validated data."""

    def _get_calculators(
        self,
        field,
        path,
        parent_path=None,
        many=False,
    ):
        """_summary_

        Args:
            field (_type_): _description_
            path (_type_): _description_
            parent_path (_type_, optional): _description_. Defaults to None.

        Returns:
            _type_: _description_
        """
        field_calculator = field.extra_kwargs.get("calculator")
        return (
            field_calculator
            and {
                str(path if not parent_path else f"{parent_path}.{path}"): many
                and field_calculator.update({"many": many})
                or field_calculator
            }
            or {}
        )

    def set_calcualtors(self):
        """_summary_

        Args:
            field_serializer (_type_): _description_
            input_type (_type_): _description_
        """
        self.calculators = {}
        for field_name, field in self.fields.items():

            if field.__class__.__name__ in ["SectionField", "SectionManyField"]:
                for sub_field in field.fields.values():
                    self.calculators.update(
                        self._get_calculators(
                            sub_field,
                            path=sub_field.field_name,
                            parent_path=field_name,
                            many=field.__class__.__name__ == "SectionManyField",
                        )
                    )
            else:
                self.calculators.update(self._get_calculators(field, path=field_name))

    def is_valid(self, *, raise_exception=False):
        """_summary_

        Args:
            raise_exception (bool, optional): _description_. Defaults to False.

        Returns:
            _type_: _description_
        """
        if self.initial_data and self.calculators:
            self.initial_data = self.calculate_data(initial_data=self.initial_data)
        return super().is_valid(raise_exception=raise_exception)

    def to_representation(self, instance):
        """_summary_

        Args:
            instance (_type_): _description_

        Returns:
            _type_: _description_
        """
        data = super().to_representation(instance)
        if self.calculators:
            data = self.calculate_data(
                initial_data=json.loads(json.dumps(data, default=json_handler))
            )
        return data

    def calculate_data(self, initial_data):
        """
        Calculates and returns processed data.

        Returns:
            dict: Processed data with calculations applied.
        """
        data = deepcopy(initial_data)
        expressionator = self.expressionator(data=data)
        calculators = deepcopy(self.calculators or {})
        temp_calculators = deepcopy(calculators)
        force_calculation = False

        while temp_calculators:
            for field_path, calculator in list(calculators.items()):
                triggered_fields = calculator.get("triggers_on", [])
                if force_calculation or not is_any_nested_item_in_list(
                    triggered_fields, calculators.keys()
                ):
                    result = expressionator.calculator(calculator, data=data)
                    data = map_calculated_data(
                        path=field_path,
                        result=result,
                        data=data,
                        many=calculator.get("many"),
                    )
                    temp_calculators.pop(field_path)

            if len(calculators) == len(temp_calculators):
                force_calculation = True
            else:
                calculators = deepcopy(temp_calculators)

        return data

    def get_fields_config(self) -> Dict[str, Any]:
        """Return the fields configuration as a dictionary.

        Returns:
            Dict[str, Any]: A dictionary representation of the fields configuration.
        """
        return get_fields_config(self.fields)
